/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */

#include "qubopt.h"

#include "QUB_BaselineSeg.h"

using namespace std;

// internal use

// don't change node->point manually
void QUB_BaselineSeg_MoveNode( QUB_BaselineSeg *seg, QUB_BaselineSeg_Node *node, int point )
{
	node->point = point;
	seg->nodes->erase( node->iter );
	(*seg->nodes)[ node->point ] = node;
	node->iter = seg->nodes->find( node->point );
}


// exported

extern "C" MIL_API QUB_BaselineSeg *       __stdcall QUB_BaselineSeg_Create( int first, int last )
{
	QUB_BaselineSeg *seg = new QUB_BaselineSeg;
	seg->first = first;
	seg->last = last;
	seg->nodes = new map<int, QUB_BaselineSeg_Node*>;

	return seg;
}

extern "C" MIL_API void                    __stdcall QUB_BaselineSeg_Free( QUB_BaselineSeg *seg )
{
	for ( map<int, QUB_BaselineSeg_Node*>::iterator ni = seg->nodes->begin(); ni != seg->nodes->end(); ++ni )
		delete ni->second;
	delete seg->nodes;
	delete seg;
}

extern "C" MIL_API int                     __stdcall QUB_BaselineSeg_GetNodeCount( QUB_BaselineSeg *seg )
{
	return seg->nodes->size();
}

extern "C" MIL_API void                    __stdcall QUB_BaselineSeg_SetFirst( QUB_BaselineSeg *seg, int first )
{
	if ( first > (seg->last + 1) )
		first = seg->last + 1;
	if ( first == seg->first )
		return;

	QUB_BaselineSeg_Node *node = QUB_BaselineSeg_GetFirstNode( seg );
	if ( node ) {
		if ( first > seg->last )
			QUB_BaselineSeg_ClearNodes( seg, seg->first, seg->last, 0 );
		else if ( first < seg->first )
			QUB_BaselineSeg_AddNode( seg, first, node->value );
		else {
			if ( seg->nodes->find( first ) == seg->nodes->end() ) {
				QUB_BaselineSeg_AddNode( seg, first, QUB_BaselineSeg_InterpolateValueAt( seg, first ) );
			}
			QUB_BaselineSeg_ClearNodes( seg, seg->first, first-1, 0 );
		}
	}

	seg->first = first;
}

extern "C" MIL_API void                    __stdcall QUB_BaselineSeg_SetLast( QUB_BaselineSeg *seg, int last )
{
	if ( last < (seg->first - 1) )
		last = seg->first - 1;
	if ( last == seg->last )
		return;

	QUB_BaselineSeg_Node *node = QUB_BaselineSeg_GetLastNode( seg );
	if ( node ) {
		if ( last < seg->first ) // empty seg
			QUB_BaselineSeg_ClearNodes( seg, seg->first, seg->last, 0 );
		else if ( last > seg->last )
			QUB_BaselineSeg_AddNode( seg, last, node->value );
		else {
			if ( seg->nodes->find( last ) == seg->nodes->end() ) {
				QUB_BaselineSeg_AddNode( seg, last, QUB_BaselineSeg_InterpolateValueAt( seg, last ) );
			}
			QUB_BaselineSeg_ClearNodes( seg, last+1, seg->last, 0 );
		}
	}

	seg->last = last;
}

extern "C" MIL_API QUB_BaselineSeg *       __stdcall QUB_BaselineSeg_ExciseFrom( QUB_BaselineSeg *seg, int first )
// split off the part starting with first as a new segment
{
	if ( (first > seg->last) || (first <= seg->first) )
		return NULL;

	QUB_BaselineSeg *newseg = QUB_BaselineSeg_Create(first, seg->last);

	if ( seg->nodes->size() ) {
		// add marginal nodes if needed
		if ( seg->nodes->find( first-1 ) == seg->nodes->end() )
			QUB_BaselineSeg_AddNode( seg, first-1, QUB_BaselineSeg_InterpolateValueAt( seg, first-1 ) );
		if ( seg->nodes->find( first ) == seg->nodes->end() )
			QUB_BaselineSeg_AddNode( seg, first, QUB_BaselineSeg_InterpolateValueAt( seg, first ) );

		QUB_BaselineSeg_Node *node = seg->nodes->find( first )->second;
		while ( node ) {
			QUB_BaselineSeg_AddNode( newseg, node->point, node->value );
			node = QUB_BaselineSeg_GetNodeAfter( seg, node );
		}
	}

	QUB_BaselineSeg_SetLast( seg, first-1 );

	return newseg;
}

extern "C" MIL_API QUB_BaselineSeg_Node *  __stdcall QUB_BaselineSeg_GetFirstNode( QUB_BaselineSeg *seg )
{
	if ( seg->nodes->size() )
		return seg->nodes->begin()->second;
	return NULL;
}

extern "C" MIL_API QUB_BaselineSeg_Node *  __stdcall QUB_BaselineSeg_GetLastNode( QUB_BaselineSeg *seg )
{
	if ( seg->nodes->size() ) {
		map<int, QUB_BaselineSeg_Node*>::iterator ni = seg->nodes->end();
		--ni;
		return ni->second;
	}
	return NULL;
}

extern "C" MIL_API QUB_BaselineSeg_Node *  __stdcall QUB_BaselineSeg_GetNodeBefore( QUB_BaselineSeg *seg, QUB_BaselineSeg_Node *node )
{
	if ( ! node )
		return QUB_BaselineSeg_GetLastNode( seg );

	if ( node->point == seg->first )
		return NULL;

	map<int, QUB_BaselineSeg_Node*>::iterator ni = node->iter;
	if ( ni == seg->nodes->begin() )
		return NULL;

	--ni;
	return ni->second;
}

extern "C" MIL_API QUB_BaselineSeg_Node *  __stdcall QUB_BaselineSeg_GetNodeAfter( QUB_BaselineSeg *seg, QUB_BaselineSeg_Node *node )
{
	if ( ! node )
		return QUB_BaselineSeg_GetFirstNode( seg );

	if ( node->point == seg->last )
		return NULL;

	map<int, QUB_BaselineSeg_Node*>::iterator ni = node->iter;
	++ni;
	if ( ni == seg->nodes->end() )
		return NULL;

	return ni->second;
}

extern "C" MIL_API int                     __stdcall QUB_BaselineSeg_InterpolateValueAt( QUB_BaselineSeg *seg, int point )
{
	QUB_BaselineSeg_Node *from, *to;
	if ( QUB_BaselineSeg_GetRangeAt( seg, point, &from, &to ) )
		return 0;
	double slope = ((double)(to->value - from->value)) / (to->point - from->point);
	double value = ( from->value + slope * (point - from->point) );
	value += (value > 0) ? 0.5 : -0.5;

	return (int) value;
}

// next three: return non-zero (negative) for error e.g. out-of-range / past-the-end
extern "C" MIL_API int                     __stdcall QUB_BaselineSeg_GetRangeAt( QUB_BaselineSeg *seg, int point, QUB_BaselineSeg_Node **from, QUB_BaselineSeg_Node **to )
{
	if ( ! seg->nodes->size() ) {
		*from = *to = NULL;
		return -1;
	}

	map<int, QUB_BaselineSeg_Node*>::iterator ni = seg->nodes->upper_bound( point );
	if ( ni == seg->nodes->end() )
		--ni;
	if ( ni == seg->nodes->begin() )
		++ni;

	*to = ni->second;
	--ni;
	*from = ni->second;

	return 0;
}

extern "C" MIL_API int                     __stdcall QUB_BaselineSeg_GetRangeBefore( QUB_BaselineSeg *seg, QUB_BaselineSeg_Node **from, QUB_BaselineSeg_Node **to )
{
	if ( ! from )
		return -1;
		
	map<int, QUB_BaselineSeg_Node*>::iterator ni = (*from)->iter;
	if ( ni == seg->nodes->begin() )
		return -1;

	*to = *from;
	--ni;
	*from = ni->second;

	return 0;
}

extern "C" MIL_API int                     __stdcall QUB_BaselineSeg_GetRangeAfter( QUB_BaselineSeg *seg, QUB_BaselineSeg_Node **from, QUB_BaselineSeg_Node **to )
{
	if ( ! to )
		return -1;

	map<int, QUB_BaselineSeg_Node*>::iterator ni = (*to)->iter;
	++ni;
	if ( ni == seg->nodes->end() )
		return -1;

	*from = *to;
	*to = ni->second;

	return 0;
}

// existing nodes at the point / in the range given will be cleared first
extern "C" MIL_API QUB_BaselineSeg_Node *  __stdcall QUB_BaselineSeg_AddNode( QUB_BaselineSeg *seg, int point, int value )
{
	QUB_BaselineSeg_ClearNodes( seg, point, point, 0 );

	QUB_BaselineSeg_Node *node = new QUB_BaselineSeg_Node;
	node->point = point;
	node->value = value;
	(*seg->nodes)[point] = node;
	node->iter = seg->nodes->find(point);

	if ( seg->nodes->find( seg->first ) == seg->nodes->end() )
		QUB_BaselineSeg_AddNode( seg, seg->first, value );
	
	if ( seg->nodes->find( seg->last ) == seg->nodes->end() )
		QUB_BaselineSeg_AddNode( seg, seg->last, value );
	else if ( QUB_BaselineSeg_GetNodeAfter(seg, node) == QUB_BaselineSeg_GetLastNode(seg) )
		QUB_BaselineSeg_GetLastNode(seg)->value = value;

	return node;
}

extern "C" MIL_API QUB_BaselineSeg_Node *  __stdcall QUB_BaselineSeg_AddNodes( QUB_BaselineSeg *seg, int count, int *points, int *values )
{
	// keep it in-bounds
	while ( count && (points[0] < seg->first) ) {
		--count;
		++points;
		++values;
	}
	while ( count && (points[count-1] > seg->last) ) {
		--count;
	}

	if ( count <= 0 )
		return NULL;

	QUB_BaselineSeg_ClearNodes( seg, points[0], points[ count-1 ], 0 );

	QUB_BaselineSeg_Node *firstNode, *node;

	for ( int i=0; i<count; ++i ) {
		node = new QUB_BaselineSeg_Node;
		node->point = points[i];
		node->value = values[i];
		(*seg->nodes)[node->point] = node;
		node->iter = seg->nodes->find(node->point);

		if ( i == 0 )
			firstNode = node;
	}

	if ( seg->nodes->find( seg->first ) == seg->nodes->end() )
		QUB_BaselineSeg_AddNode( seg, seg->first, values[0] );
	if ( seg->nodes->find( seg->last ) == seg->nodes->end() )
		QUB_BaselineSeg_AddNode( seg, seg->last, values[ count-1 ] );
	else if ( QUB_BaselineSeg_GetNodeAfter(seg, node) == QUB_BaselineSeg_GetLastNode(seg) )
		QUB_BaselineSeg_GetLastNode(seg)->value = values[ count-1 ];

	return firstNode;
}

extern "C" MIL_API void                    __stdcall QUB_BaselineSeg_ClearNodes( QUB_BaselineSeg *seg, int first, int last, int repairFirstLast )
{
	map<int, QUB_BaselineSeg_Node*>::iterator ni = seg->nodes->lower_bound( first );
	if ( ni == seg->nodes->end() )
		return;

	QUB_BaselineSeg_Node *node = ni->second;
	QUB_BaselineSeg_Node *next = QUB_BaselineSeg_GetNodeAfter( seg, node );
	while ( next && (node->point <= last) ) {
		seg->nodes->erase( node->iter );
		delete node;
		node = next;
		next = QUB_BaselineSeg_GetNodeAfter( seg, node );
	}
	if ( node->point <= last ) {
		seg->nodes->erase( node->iter );
		delete node;
	}

	if ( repairFirstLast && seg->nodes->size() ) {
		if ( seg->nodes->find( seg->first ) == seg->nodes->end() )
			QUB_BaselineSeg_AddNode( seg, seg->first, seg->nodes->begin()->second->value );
		if ( seg->nodes->find( seg->last ) == seg->nodes->end() ) {
			ni = seg->nodes->end();
			--ni;
			QUB_BaselineSeg_AddNode( seg, seg->last, ni->second->value );
		}
	}
}

extern "C" MIL_API void                    __stdcall QUB_BaselineSeg_Offset( QUB_BaselineSeg *seg, int points )
{
	seg->first += points;
	seg->last += points;

	vector<QUB_BaselineSeg_Node*> nodes;

	for ( map<int, QUB_BaselineSeg_Node*>::iterator ni = seg->nodes->begin(); ni != seg->nodes->end(); ++ni )
		nodes.push_back( ni->second );
	seg->nodes->clear();

	for ( int i=0; i<int(nodes.size()); ++i ) {
		nodes[i]->point += points;
		(*seg->nodes)[ nodes[i]->point ] = nodes[i];
		nodes[i]->iter = seg->nodes->find(nodes[i]->point);
	}
}

extern "C" MIL_API void                    __stdcall QUB_BaselineSeg_Insert( QUB_BaselineSeg *seg, int before, int points )
{
	if ( before < seg->first || before > (seg->last + 1) )
		return;

	seg->last += points;

	if ( ! seg->nodes->size() )
		return;

	if ( before > seg->last ) {
		map<int, QUB_BaselineSeg_Node*>::iterator ni = seg->nodes->end();
		--ni;
		QUB_BaselineSeg_AddNode( seg, seg->last, ni->second->value );

		return;
	}

	QUB_BaselineSeg_Node *node = QUB_BaselineSeg_GetLastNode( seg );
	while ( node && node->point >= before ) {
		QUB_BaselineSeg_MoveNode( seg, node, node->point + points );
		node = QUB_BaselineSeg_GetNodeBefore( seg, node );
	}

	if ( ! node ) // we just moved the first node up, so add a new first node
		QUB_BaselineSeg_AddNode( seg, seg->first, seg->nodes->begin()->second->value );
}


